/*
 * FactoryLibrary_ltdl.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file FactoryLibrary_ltdl.hh
 *
 * Header file for the FactoryLibrary_ltdl class.
 */

#ifndef _factory_library_impl_hh
#define _factory_library_impl_hh

#include <map>
#include <string.h>

#include "AllupS.h"

#include "HeyParser.hh"

#include "factory_library.h"
#include "ltdl.h"

using namespace std;

/**
 * An edu::utah::pces::FactoryLibrary implementation that uses libltdl.
 *
 * @sa AllupImpl
 * @sa libtool
 */
class FactoryLibrary_ltdl : public virtual POA_edu::utah::pces::FactoryLibrary
{

public:

    /**
     * Open a library with the given name.
     *
     * @param poa The POA to use when activating any new objects.
     * @param name The name of the library.
     * @return A CORBA object reference to the new FactoryLibrary_ltdl object.
     *
     * @exception edu::utah::pces::NoSuchLibrary if the library could not be
     * found or loaded.
     */
    static edu::utah::pces::FactoryLibrary_ptr OpenLibrary(
	PortableServer::POA_ptr poa,
	const char *name)
	throw (edu::utah::pces::NoSuchLibrary,
	       CORBA::SystemException);

    /**
     * Shutdown the library interface.
     */
    static void Shutdown(void);

    /**
     * Deconstruct a FactoryLibrary_ltdl object.
     */
    virtual ~FactoryLibrary_ltdl(void);

    /** @copydoc edu::utah::pces::FactoryLibrary::Name */
    virtual char *Name(void)
	throw (CORBA::SystemException);

    /** @copydoc edu::utah::pces::FactoryLibrary::Hey */
    virtual CORBA::Long Hey(const edu::utah::pces::ArgV &args,
			    CORBA::String_out o,
			    CORBA::String_out e)
	throw (CORBA::SystemException);

    /**
     * @exception CORBA::BAD_INV_ORDER when called.  This object should always
     * be activated under a single threaded POA.
     */
    virtual PortableServer::POA_ptr _default_POA(void)
	throw (CORBA::SystemException)
    {
	throw CORBA::BAD_INV_ORDER();
    };

protected:
    
    /**
     * Construct a FactoryLibrary_ltdl object with the given arguments.
     *
     * @param poa The POA this object was activated under and any new objects
     * should be activated under.
     * @param name The name of this shared library.
     * @param dlh The libtool handle on the shared library.
     * @param fm The library's factory_method_t.
     */
    FactoryLibrary_ltdl(PortableServer::POA_ptr poa,
			const char *name,
			lt_dlhandle dlh,
			factory_method_t fm);
    
private:

    /**
     * Map of library names to FactoryLibrary_ltdl objects.
     */
    static map<string, FactoryLibrary_ltdl *> open_libraries;

    /**
     * Cached reference to the ORB object.
     */
    static CORBA::ORB_var orb;

    /**
     * Construct a FactoryLibrary_ltdl object with the given values.
     *
     * @param name The name of the library.
     * @param library The libtool handle on the library.
     * @param fm The reference to the library's factory_method_t function.
     */
    FactoryLibrary_ltdl(const char *name,
			lt_dlhandle library,
			factory_method_t fm);

    /**
     * The POA this object was activated under and any new objects should be
     * activated under.
     */
    PortableServer::POA_var fl_POA;
    
    /**
     * The name of the library being managed by this object.
     */
    CORBA::String_var fl_Name;

    /**
     * The number of times this library has been opened.
     */
    unsigned int fl_OpenCount;

    /**
     * Libtool handle on the shared library.
     */
    lt_dlhandle fl_Library;

    /**
     * Cached reference to the factory_method_t in the shared library.
     */
    factory_method_t fl_Method;
    
};

#endif
