/*
 * StrictPolicyDelegateImpl.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file StrictPolicyDelegateImpl.hh
 *
 * Header file for the StrictPolicyDelegateImpl class.
 */

#ifndef _strict_policy_delegate_impl_hh
#define _strict_policy_delegate_impl_hh

#include "DelegateS.h"
#include "DelegateMixin.hh"

/**
 * Base class for objects that wish to act as a delegate of a
 * BrokerPolicies::StrictPolicy object.
 *
 * @param T The main object type.
 * @param T_var The object's _var type.
 */
class StrictPolicyDelegateImpl :
    public virtual DelegateMixin<BrokerPolicies::StrictPolicy,
				 BrokerPolicies::StrictPolicy_var>,
    public virtual POA_BrokerDelegates::StrictPolicyDelegate
{

public:

    /**
     * Construct a StrictPolicyDelegateImpl object.
     */
    StrictPolicyDelegateImpl(void)
    {
    };

    /**
     * Deconstruct a StrictPolicyDelegateImpl object.
     */
    virtual ~StrictPolicyDelegateImpl(void)
    {
    };

    /** @copydoc Broker::Policy::Name */
    virtual char *Name(void)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	return this->dm_RemoteObject->Name();
    };
    
    /** @copydoc Broker::Policy::AddTask */
    virtual void AddTask(Broker::Task_ptr new_task,
			 const Broker::ScheduleParameters &sp)
	throw (CORBA::SystemException,
	       Broker::DuplicateScheduleParameter,
	       Broker::InvalidScheduleParameter,
	       Broker::MissingScheduleParameter)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->AddTask(new_task, sp);
    };
    
    /** @copydoc Broker::Policy::RemoveTask */
    virtual void RemoveTask(Broker::Task_ptr added_task)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->RemoveTask(added_task);
    };
    
    /** @copydoc Broker::Policy::Activate */
    virtual void Activate(const Broker::TaskList &tasks)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->Activate(tasks);
    };
    
    /** @copydoc Broker::Policy::Deactivate */
    virtual void Deactivate(void)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->Deactivate();
    };
    
    /** @copydoc Broker::Policy::ChangeTaskCPU */
    virtual Broker::CPUReserve ChangeTaskCPU(Broker::RealTimeTask_ptr task,
					     const Broker::CPUReserve &advice)
	throw (CORBA::SystemException, Broker::InvalidState)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	return this->dm_RemoteObject->ChangeTaskCPU(task, advice);
    };

    /** @copydoc Broker::Policy::GetMaxCPUAllocation */
    virtual CORBA::Float GetMaxCPUAllocation(void)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	return this->dm_RemoteObject->GetMaxCPUAllocation();
    };
    
    /** @copydoc Broker::Policy::SetMaxCPUAllocation */
    virtual void SetMaxCPUAllocation(CORBA::Float amount)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->SetMaxCPUAllocation(amount);
    };

    /** @copydoc Broker::Policy::GetTaskList */
    virtual Broker::TaskList *GetTaskList(void)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	return this->dm_RemoteObject->GetTaskList();
    };

    /** @copydoc BrokerPolicies::StrictPolicy::SetTaskPriority */
    virtual void SetTaskPriority(Broker::Task_ptr task, CORBA::Short priority)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	this->dm_RemoteObject->SetTaskPriority(task, priority);
    };

    /** @copydoc BrokerPolicies::StrictPolicy::GetTaskPriority */
    virtual CORBA::Short GetTaskPriority(Broker::Task_ptr task)
	throw (CORBA::SystemException)
    {
	if( CORBA::is_nil(this->dm_RemoteObject.in()) )
	{
	    throw CORBA::BAD_INV_ORDER();
	}
	return this->dm_RemoteObject->GetTaskPriority(task);
    };

};

#endif
