/*
 * factory_library.hh
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file factory_library.h
 *
 * Header file that presents a common interface for shared libraries.  The
 * interface is basically a single function, flFactoryMethodV, that can perform
 * a variety of tasks for the caller.
 *
 * @see cbhey.cc
 * @see BrokerFactory.cc
 * @see PolicyFactory.cc
 * @see DelegateFactory.cc
 */

#ifndef _factory_library_h
#define _factory_library_h

#ifdef __cplusplus
extern "C" {
#endif

#include <stdarg.h>

/**
 * Sub-operations for the primary library function below.
 */
typedef enum {
    FLO_MIN,	/**< The minimum legal value. */
    FLO_QUERY,	/**< Ask the library if it contains support for something. */
    FLO_HEY,	/**< Ask the library to handle a 'hey' request. */
    FLO_MAX	/**< The maximum legal value. */
} factory_library_op_t;

/**
 * Argument tags for the factory_method_t functions.
 */
enum {
    FMA_TAG_DONE,	/**< () - Terminate the tag list. */
    FMA_ORB,		/**< (CORBA::ORB_ptr) - Pointer to an ORB. */
    FMA_POA,		/**< (PortableServer::POA_ptr) - Pointer to a POA. */
    FMA_HeyParser,	/**< (HeyParser *) - Pointer to an active HeyParser. */
    FMA_HeyServerHint,	/**< (const char *) - Hint for the type of server. */
    FMA_StandardOut,	/**< (ostream *) - Standard output stream. */
    FMA_StandardError,	/**< (ostream *) - Standard error stream. */
    FMA_USER_BASE = 0x1000,	/**< Base for user tags. */
};

/**
 * Function type for the primary library function interface.
 *
 * @param op The sub-operation to execute.
 * @param tag The first tag in the tag list.
 * @param args The remaining tags in the tag list.
 * @return Zero if the operation succeeded, an errno value otherwise.
 */
typedef int (*factory_method_t)(factory_library_op_t op,
				int tag,
				va_list args);

/**
 * The symbol name for the primary library function.
 */
#define FACTORY_METHOD_SYMBOL flFactoryMethodV

/**
 * The primary library function.  This function provides a generic interface
 * to the capabilities of a shared library.
 *
 * @param op The sub-operation to execute.
 * @param tag The first tag in the tag list.
 * @param args The remaining tags in the tag list.
 * @return Zero if the operation succeeded, an errno value otherwise.
 */
int FACTORY_METHOD_SYMBOL(factory_library_op_t op,
			  int tag,
			  va_list args);

/**
 * Wrapper function that calls a given factory_method_t with the arguments
 * given on the stack.
 *
 * @param fm The primary library function to pass the request to.
 * @param op The sub-operation to execute.
 * @param tag The first tag in the tag list.
 * @return Zero if the operation succeeded, an errno value otherwise.
 */
int flFactoryMethod(factory_method_t fm,
		    factory_library_op_t op,
		    int tag,
		    ...);

#ifdef __cplusplus
}
#endif

#endif
